package com.sun.jndi.cosnaming;

import java.util.Enumeration;
import javax.naming.CompoundName;
import java.util.Vector;
import javax.naming.CompositeName;
import javax.naming.InvalidNameException;
import org.omg.CosNaming.NameComponent;
import javax.naming.NamingException;
import javax.naming.Name;
import java.util.Properties;
import javax.naming.NameParser;

public final class CNNameParser implements NameParser
{
    private static final Properties mySyntax;
    private static final char kindSeparator = '.';
    private static final char compSeparator = '/';
    private static final char escapeChar = '\\';
    
    public CNNameParser() {
    }
    
    @Override
    public Name parse(String s) throws NamingException {
        return new CNNameParser.CNCompoundName(insStringToStringifiedComps(s).elements());
    }
    
    static NameComponent[] nameToCosName(Name name) throws InvalidNameException {
        int size = name.size();
        if (size == 0) {
            return new NameComponent[0];
        }
        NameComponent[] array = new NameComponent[size];
        for (int i = 0; i < size; ++i) {
            array[i] = parseComponent(name.get(i));
        }
        return array;
    }
    
    static String cosNameToInsString(NameComponent[] array) {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < array.length; ++i) {
            if (i > 0) {
                sb.append('/');
            }
            sb.append(stringifyComponent(array[i]));
        }
        return sb.toString();
    }
    
    static Name cosNameToName(NameComponent[] array) {
        CompositeName compositeName = new CompositeName();
        for (int n = 0; array != null && n < array.length; ++n) {
            try {
                compositeName.add(stringifyComponent(array[n]));
            }
            catch (InvalidNameException ex) {}
        }
        return compositeName;
    }
    
    private static Vector<String> insStringToStringifiedComps(String s) throws InvalidNameException {
        int length = s.length();
        Vector vector = new Vector(10);
        char[] value = new char[length];
        char[] value2 = new char[length];
        int i = 0;
        while (i < length) {
            int count2;
            int count = count2 = 0;
            int n = 1;
            while (i < length && s.charAt(i) != '/') {
                if (s.charAt(i) == '\\') {
                    if (i + 1 >= length) {
                        throw new InvalidNameException(s + ": unescaped \\ at end of component");
                    }
                    if (isMeta(s.charAt(i + 1))) {
                        ++i;
                        if (n != 0) {
                            value[count2++] = s.charAt(i++);
                        }
                        else {
                            value2[count++] = s.charAt(i++);
                        }
                    }
                    else {
                        throw new InvalidNameException(s + ": invalid character being escaped");
                    }
                }
                else if (n != 0 && s.charAt(i) == '.') {
                    ++i;
                    n = 0;
                }
                else if (n != 0) {
                    value[count2++] = s.charAt(i++);
                }
                else {
                    value2[count++] = s.charAt(i++);
                }
            }
            vector.addElement(stringifyComponent(new NameComponent(new String(value, 0, count2), new String(value2, 0, count))));
            if (i >= length) {
                continue;
            }
            ++i;
        }
        return vector;
    }
    
    private static NameComponent parseComponent(String s) throws InvalidNameException {
        NameComponent nameComponent = new NameComponent();
        int n = -1;
        int length = s.length();
        int count = 0;
        char[] array = new char[length];
        int n2 = 0;
        for (int n3 = 0; n3 < length && n < 0; ++n3) {
            if (n2 != 0) {
                array[count++] = s.charAt(n3);
                n2 = 0;
            }
            else if (s.charAt(n3) == '\\') {
                if (n3 + 1 >= length) {
                    throw new InvalidNameException(s + ": unescaped \\ at end of component");
                }
                if (isMeta(s.charAt(n3 + 1))) {
                    n2 = 1;
                }
                else {
                    throw new InvalidNameException(s + ": invalid character being escaped");
                }
            }
            else if (s.charAt(n3) == '.') {
                n = n3;
            }
            else {
                array[count++] = s.charAt(n3);
            }
        }
        nameComponent.id = new String(array, 0, count);
        if (n < 0) {
            nameComponent.kind = "";
            return nameComponent;
        }
        int count2 = 0;
        int n4 = 0;
        for (int i = n + 1; i < length; ++i) {
            if (n4 != 0) {
                array[count2++] = s.charAt(i);
                n4 = 0;
            }
            else if (s.charAt(i) == '\\') {
                if (i + 1 >= length) {
                    throw new InvalidNameException(s + ": unescaped \\ at end of component");
                }
                if (isMeta(s.charAt(i + 1))) {
                    n4 = 1;
                }
                else {
                    throw new InvalidNameException(s + ": invalid character being escaped");
                }
            }
            else {
                array[count2++] = s.charAt(i);
            }
        }
        nameComponent.kind = new String(array, 0, count2);
        return nameComponent;
    }
    
    private static String stringifyComponent(NameComponent nameComponent) {
        StringBuffer sb = new StringBuffer(escape(nameComponent.id));
        if (nameComponent.kind != null && !nameComponent.kind.equals("")) {
            sb.append(46 + escape(nameComponent.kind));
        }
        if (sb.length() == 0) {
            return ".";
        }
        return sb.toString();
    }
    
    private static String escape(String s) {
        if (s.indexOf(46) < 0 && s.indexOf(47) < 0 && s.indexOf(92) < 0) {
            return s;
        }
        int length = s.length();
        int count = 0;
        char[] value = new char[length + length];
        for (int i = 0; i < length; value[count++] = s.charAt(i), ++i) {
            if (isMeta(s.charAt(i))) {
                value[count++] = '\\';
            }
        }
        return new String(value, 0, count);
    }
    
    private static boolean isMeta(char c) {
        switch (c) {
            case '.':
            case '/':
            case '\\': {
                return true;
            }
            default: {
                return false;
            }
        }
    }
    
    static {
        (mySyntax = new Properties()).put("jndi.syntax.direction", "left_to_right");
        CNNameParser.mySyntax.put("jndi.syntax.separator", "/");
        CNNameParser.mySyntax.put("jndi.syntax.escape", "\\");
    }
    
    static final class CNCompoundName extends CompoundName
    {
        private static final long serialVersionUID = -6599252802678482317L;
        
        CNCompoundName(Enumeration<String> comps) {
            super(comps, CNNameParser.mySyntax);
        }
        
        @Override
        public Object clone() {
            return new CNNameParser.CNCompoundName(this.getAll());
        }
        
        @Override
        public Name getPrefix(int posn) {
            return new CNNameParser.CNCompoundName(super.getPrefix(posn).getAll());
        }
        
        @Override
        public Name getSuffix(int posn) {
            return new CNNameParser.CNCompoundName(super.getSuffix(posn).getAll());
        }
        
        @Override
        public String toString() {
            try {
                return CNNameParser.cosNameToInsString(CNNameParser.nameToCosName(this));
            }
            catch (InvalidNameException ex) {
                return super.toString();
            }
        }
    }
}
